from __future__ import absolute_import, division, print_function

__all__ = (
    "DiffsFoundError",
    "EyesError",
    "NewTestError",
    "OutOfBoundsError",
    "TestFailedError",
    "USDKFailure",
)


class EyesError(Exception):
    """
    Applitools Eyes Exception.
    """


class OutOfBoundsError(EyesError):
    """
    Indicates that an element is outside a specific boundary (e.g, region outside a frame,
    or point outside an image).
    """


class EyesServiceUnavailableError(EyesError):
    """Indicates that Eyes concurrency limit is exceeded"""


class TestFailedError(Exception):
    """
    Indicates that a test did not pass (i.e., test either failed or is a new test).
    """

    __test__ = False  # avoid warnings in test frameworks

    def __init__(
        self, test_results=None, scenario_id_or_name=None, app_id_or_name=None
    ):
        if all([test_results, scenario_id_or_name, app_id_or_name]):
            msg = "'{}' of '{}'. See details at {}".format(
                scenario_id_or_name, app_id_or_name, test_results.url
            )
            self._test_results = test_results
        else:
            msg = test_results
        super(TestFailedError, self).__init__(msg)


class NewTestError(TestFailedError):
    """
    Indicates that a test is a new test.
    """


class DiffsFoundError(TestFailedError):
    """
    Indicates that an existing test ended, and that differences where found from the baseline.
    """


class USDKFailure(EyesError):
    """
    Generic error raised by Universal SDK
    """

    def __str__(self):
        """Human-readable error representation."""
        return "\n".join(self.args)



# GENERATED FILE #
from enum import Enum

from applitools.common.utils.general_utils import DeprecatedEnumVariant

__all__ = ("AccessibilityGuidelinesVersion",)


class AccessibilityGuidelinesVersion(Enum):
    WCAG_2_0 = "WCAG_2_0"
    WCAG_2_1 = "WCAG_2_1"



# GENERATED FILE #
from enum import Enum

from applitools.common.utils.general_utils import DeprecatedEnumVariant

__all__ = ("AccessibilityStatus",)


class AccessibilityStatus(Enum):
    """
    Accessibility status.
    """

    # Session has passed accessibility validation.
    Passed = "Passed"
    # Session hasn't passed accessibility validation.
    Failed = "Failed"



from __future__ import absolute_import, division, print_function, unicode_literals

from ..common import deprecated
from ..common.extract_text import OCRRegion, TextRegion, TextRegionSettings

deprecated.module(__name__)
__all__ = "OCRRegion", "TextRegion", "TextRegionSettings"



from __future__ import absolute_import, division, print_function, unicode_literals

from ..common import deprecated
from ..common.triggers import MouseTrigger, TextTrigger

deprecated.module(__name__)
__all__ = "MouseTrigger", "TextTrigger"



from __future__ import absolute_import, division, print_function

import os
from functools import wraps
from typing import TYPE_CHECKING
from urllib.parse import urlparse
from warnings import warn

"""
General purpose utilities.
"""

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Optional, Text


def cached_property(f):
    # type: (Callable) -> Any
    """
    Returns a cached property that is calculated by function f
    """

    def get(self):
        try:
            return self._property_cache[f]
        except AttributeError:
            self._property_cache = {}
            x = self._property_cache[f] = f(self)
            return x
        except KeyError:
            x = self._property_cache[f] = f(self)
            return x

    return property(get)


def get_env_with_prefix(env_name, default=None):
    # type: (Text, Optional[Text]) -> Optional[Text]
    """
    Takes name of ENV variable, check if exists origin and with list of prefixes
    """
    prefixes_to_check = ["bamboo"]
    try:
        return os.environ[env_name]
    except KeyError:
        for prefix in prefixes_to_check:
            name = "{}_{}".format(prefix, env_name)
            value = os.getenv(name)
            if value:
                return value
    return default


class DeprecatedEnumVariant(object):
    """
    Deprecate Enum variant with message from docstring
    """

    def __init__(self, fget=None, msg=None):
        self.fget = fget
        self.__doc__ = fget.__doc__

    def __get__(self, instance, ownerclass=None):
        warn(self.__doc__, DeprecationWarning, stacklevel=2)
        return self.fget(ownerclass)


class CustomEnumMethod:
    def __init__(self, func):
        self.func = func
        self.__doc__ = func.__doc__

    def __get__(self, obj, objtype=None):
        @wraps(self.func)
        def wrapper(*args, **kwargs):
            if not args or not isinstance(args[0], str):
                raise ValueError("Custom method must be called with a string pattern")
            return self.func(objtype, *args, **kwargs)

        return wrapper


def is_region_like_dict(obj):
    # type: (Dict) -> bool
    """
    Checks if the given object is a dictionary with keys corresponding to a Region.

    Args:
    obj: Object to check.

    Returns:
    bool: True if obj is a dictionary with Region keys, False otherwise.
    """
    return isinstance(obj, dict) and set(obj.keys()) == {
        "left",
        "top",
        "width",
        "height",
    }


def deep_compare(obj1, obj2):
    # type: (Any, Any) -> bool
    """
    Compares two objects deeply.

    Args:
    obj1: First object to compare.
    obj2: Second object to compare.

    Returns:
    bool: True if obj1 and obj2 are the same, False otherwise.
    """
    from applitools.common.geometry import Region

    if is_region_like_dict(obj1):
        obj1 = Region.from_(obj1)

    if is_region_like_dict(obj2):
        obj2 = Region.from_(obj2)

    if type(obj1) is not type(obj2):
        if isinstance(obj1, float) and obj1 == float(obj2):
            return True
        return False

    if isinstance(obj1, dict):
        if obj1.keys() != obj2.keys():
            return False
        return all(deep_compare(obj1[key], obj2[key]) for key in obj1)

    if isinstance(obj1, list):
        if len(obj1) != len(obj2):
            return False
        return all(deep_compare(item1, item2) for item1, item2 in zip(obj1, obj2))

    return obj1 == obj2


def is_url(string):
    try:
        result = urlparse(string)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False



from __future__ import absolute_import, division, print_function

from copy import copy
from typing import TYPE_CHECKING, Dict, List, Optional, Text, TypeVar, Union

import attr

from .fluent.images_check_settings import image_path_or_bytes
from .fluent.target_path import TargetPath
from .geometry import Rectangle, Region
from .utils import argument_guard
from .validators import is_list_or_tuple, is_webelement

if TYPE_CHECKING:
    from .fluent.target_path import RegionLocator
    from .optional_deps import Image, WebElement
    from .utils.custom_types import BySelector


class TextRegionSettings(object):
    def __init__(self, *patterns):
        # type: (*Union[Text, List[Text]]) -> None
        self._first_only = None  # type: Optional[bool]
        self._language = "eng"  # type: Text
        self._ignore_case = None  # type: Optional[bool]
        self._patterns = []  # type: List[Text]
        self._image = None  # type: Union[None, Image, Text]
        self._add_patterns(*patterns)

    @property
    def is_first_only(self):
        # type: () -> bool
        return self._first_only

    def image(self, image):
        # type: (Union[Image, Text]) -> TextRegionSettings
        """
        Perform text search on a given image
        """
        cloned = self._clone()
        cloned._image = image_path_or_bytes(image)
        return cloned

    def _add_patterns(self, *patterns):
        if not patterns:
            return self
        if len(patterns) == 1 and argument_guard.is_list_or_tuple(patterns[0]):
            patterns = patterns[0]
        argument_guard.is_list_or_tuple(patterns)
        cloned = self._clone()
        cloned._patterns.extend(list(patterns))
        return cloned

    def ignore_case(self, ignore=True):
        # type: (bool) -> TextRegionSettings
        cloned = self._clone()
        cloned._ignore_case = ignore
        return cloned

    def first_only(self):
        cloned = self._clone()
        cloned._first_only = True
        return cloned

    def language(self, language):
        # type: (Text) -> TextRegionSettings
        argument_guard.is_a(language, str)
        assert language not in ["eng"], "Unsupported language"
        cloned = self._clone()
        cloned._language = language
        return cloned

    def _clone(self):
        # type: () -> TextRegionSettings
        cloned = copy(self)
        cloned._patterns = copy(cloned._patterns)
        return self


@attr.s(slots=True, init=False)
class TextRegion(Rectangle):
    left = attr.ib()  # type: int
    top = attr.ib()  # type: int
    width = attr.ib()  # type: int
    height = attr.ib()  # type: int
    text = attr.ib()  # type: Text

    @property
    def x(self):
        return self.left

    @property
    def y(self):
        return self.top

    def __init__(
        self,
        left,  # type: int
        top,  # type: int
        width,  # type: int
        height,  # type: int
        text,  # type: Text
    ):
        # type: (...) -> None
        super(TextRegion, self).__init__(left, top, width, height)
        self.text = text


class OCRRegion(object):
    def __init__(self, target):
        # type: (Union[Region, Text, BySelector, WebElement, RegionLocator]) -> None
        if isinstance(target, str) or is_webelement(target):
            self.target = TargetPath.region(target)
        elif is_list_or_tuple(target):
            self.target = TargetPath.region(*target)
        else:
            self.target = target
        self._hint = None  # type: Optional[Text]
        self._language = "eng"  # type: Optional[Text]
        self._min_match = None  # type: Optional[float]

    def min_match(self, min_match):
        # type: (float) -> Self
        cloned = self._clone()
        cloned._min_match = min_match
        return cloned

    def hint(self, hint):
        # type: (Text) -> Self
        cloned = self._clone()
        cloned._hint = hint
        return cloned

    def language(self, language):
        # type: (Text) -> Self
        cloned = self._clone()
        cloned._language = language
        return cloned

    def _clone(self):
        # type: () -> Self
        return copy(self)


class ImageOCRRegion(OCRRegion):
    def __init__(self, image, region_in_image=None):
        # type: (Union[Image, Text], Optional[Region]) -> None
        super(ImageOCRRegion, self).__init__(region_in_image)
        self.image = image_path_or_bytes(image)


PATTERN_TEXT_REGIONS = Dict[Text, List[TextRegion]]  # typedef
Self = TypeVar("Self", bound="OCRRegion")  # typedef



from __future__ import absolute_import, division, print_function

import atexit
import logging
import weakref
from json import dumps, loads
from queue import Empty, Queue
from threading import Thread
from time import time
from typing import TYPE_CHECKING

from websocket import WebSocket

from applitools.core_universal import get_instance

from .object_registry import RefId

if TYPE_CHECKING:
    from typing import Optional, Text

    from applitools.core_universal.server import SDKServer

    from .command_context import CommandContext

_all_sockets = []
_logger = logging.getLogger(__name__)

try:
    from builtins import TimeoutError
except ImportError:

    class TimeoutError(Exception):
        pass


class USDKConnection(object):
    def __init__(self, websocket):
        # type: (WebSocket) -> None
        self._websocket = websocket
        self._response_queues = {}
        weak_socket = weakref.ref(self._websocket)
        self._receiver_thread = Thread(
            target=self._receiver_loop,
            name="USDK Receiver",
            args=(weak_socket, self._response_queues),
        )
        # Receiver threads are designed to exit even if they serve leaked unclosed
        # connections. But non-daemon threads are joined on shutdown deadlocking with
        # garbage collector and atexit manager that should close the socket
        self._receiver_thread.daemon = True
        _all_sockets.append(weak_socket)
        self._receiver_thread.start()

    @classmethod
    def create(cls, server=None, mask_log=None):
        # type: (Optional[SDKServer], Optional[bool]) -> USDKConnection
        server = server or get_instance(mask_log=mask_log)
        websocket = WebSocket()
        websocket.connect("ws://localhost:{}/eyes".format(server.port))
        return cls(websocket)

    def notification(self, name, payload):
        # type: (Text, dict) -> None
        self._websocket.send(dumps({"name": name, "payload": payload}))

    def command(self, context, name, payload, wait_timeout):
        # type: (CommandContext, Text, dict, float) -> Optional[dict]
        assert context.key.object_registry_id not in self._response_queues
        queue = Queue()
        self._response_queues[context.key.object_registry_id] = queue
        self._websocket.send(
            dumps({"name": name, "key": str(context.key), "payload": payload})
        )
        deadline = time() + wait_timeout
        while True:
            try:
                event, obj = queue.get(timeout=deadline - time())
            except Empty:
                raise TimeoutError("USDK Command timeout")
            if event == "result":
                return obj
            elif event == "callback":
                context.execute_callback(obj)
            else:
                raise RuntimeError("Unexpected", event, obj)

    def response(self, key, name, result=None, error=None):
        # type: (Text, Text, Optional[dict], Optional[dict]) -> None
        payload = {}
        if result is not None:
            payload["result"] = result
        elif error is not None:
            payload["error"] = error
        data = {"key": key, "name": name, "payload": payload}
        self._websocket.send(dumps(data))

    def close(self):
        if self._websocket:
            self._websocket.close()
            self._websocket = None
            self._receiver_thread = None

    def __del__(self):
        self.close()

    @staticmethod
    def _receiver_loop(weak_socket, response_queues):
        while True:
            try:
                socket = weak_socket()
                if not socket:
                    raise EOFError
                response = socket.recv()
                del socket
                if not response:
                    raise EOFError
                response = loads(response)
                command_name = response["name"]
                if command_name.startswith("Driver."):  # specdriver callback
                    key = command_key(response)
                    response_queues[key.object_registry_id].put(("callback", response))
                elif command_name == "Logger.log":
                    entry = response["payload"]
                    level = logging.getLevelName(entry["level"].upper())
                    _logger.log(level, entry["message"])
                else:
                    key = RefId.from_str(response["key"])
                    response_queues.pop(key.object_registry_id).put(
                        ("result", response)
                    )
            except Exception as exc:
                socket = weak_socket()
                if socket:
                    socket.abort()
                for queue in response_queues.values():
                    queue.put(("exception", exc))
                break


def command_key(response):
    # type: (dict) -> RefId
    payload = response["payload"]
    if "context" in payload:
        obj_id = payload["context"]["applitools-ref-id"]
    else:
        obj_id = payload["driver"]["applitools-ref-id"]
    return RefId.from_str(obj_id)


@atexit.register
def ensure_all_closed():
    for weak_ref in _all_sockets:
        socket = weak_ref()
        if socket:
            socket.close()



from __future__ import absolute_import, division, print_function

from typing import TYPE_CHECKING

from .context_vars import set_object_registry
from .schema import (
    CheckSettings,
    CloseBatchSettings,
    CloseSettings,
    DeleteTestSettings,
    ECClientSettings,
    ExtractTextSettings,
    EyesConfig,
    ImageTarget,
    LocateSettings,
    OCRSearchSettings,
    OpenSettings,
    Size,
)

if TYPE_CHECKING:
    from typing import List, Tuple

    from .. import common
    from . import config as cfg
    from . import ec_client_settings, extract_text, locators
    from . import target as t
    from .batch_close import _EnabledBatchClose  # noqa
    from .fluent import selenium_check_settings as cs
    from .object_registry import ObjectRegistry
    from .optional_deps import WebDriver
    from .utils.custom_types import ViewPort


class Marshaller(object):
    def __init__(self, object_registry):
        # type: (ObjectRegistry) -> None
        self._object_registry = object_registry

    def marshal_viewport_size(self, viewport_size):
        # type: (ViewPort) -> dict
        with set_object_registry(self._object_registry):
            return Size().dump(viewport_size)

    def marshal_webdriver_ref(self, driver):
        # type: (WebDriver) -> dict
        return self._object_registry.marshal_driver(driver)

    def marshal_ec_client_settings(self, client_settings):
        # type: (ec_client_settings.ECClientSettings) -> dict
        return ECClientSettings().dump(client_settings)

    def marshal_enabled_batch_close(self, close_batches):
        # type: (_EnabledBatchClose) -> dict
        with set_object_registry(self._object_registry):
            return CloseBatchSettings().dump(close_batches)

    def marshal_delete_test_settings(self, test_results):
        # type: (common.TestResults) -> dict
        with set_object_registry(self._object_registry):
            return DeleteTestSettings().dump(test_results)

    def marshal_configuration(self, configuration):
        # type: (cfg.Configuration) -> dict
        with set_object_registry(self._object_registry):
            open = OpenSettings().dump(configuration)
            config = EyesConfig().dump(configuration)
            close = CloseSettings().dump(configuration)
            return {"open": open, "screenshot": config, "check": config, "close": close}

    def marshal_check_settings(self, check_settings):
        # type: (cs.SeleniumCheckSettings) -> dict
        with set_object_registry(self._object_registry):
            return CheckSettings().dump(check_settings.values)

    def marshal_image_target(self, image_target):
        # type: (t.ImageTarget) -> dict
        with set_object_registry(self._object_registry):
            return ImageTarget().dump(image_target)

    def marshal_locate_settings(self, locate_settings):
        # type: (locators.VisualLocatorSettings) -> dict
        with set_object_registry(self._object_registry):
            return LocateSettings().dump(locate_settings.values)

    def marshal_ocr_extract_settings(self, extract_settings):
        # type: (Tuple[extract_text.OCRRegion, ...]) -> List[dict]
        with set_object_registry(self._object_registry):
            return [ExtractTextSettings().dump(s) for s in extract_settings]

    def marshal_ocr_search_settings(self, search_settings):
        # type: (extract_text.TextRegionSettings) -> dict
        with set_object_registry(self._object_registry):
            return OCRSearchSettings().dump(search_settings)



from __future__ import absolute_import, division, print_function, unicode_literals

from ..common import deprecated
from ..common.batch_close import BatchClose

deprecated.module(__name__)
__all__ = ("BatchClose",)



# GENERATED FILE #
from enum import Enum

from applitools.common.utils.general_utils import DeprecatedEnumVariant

__all__ = ("AccessibilityLevel",)


class AccessibilityLevel(Enum):
    AA = "AA"
    AAA = "AAA"



from __future__ import absolute_import, division, print_function

try:
    from os import PathLike, fspath
except ImportError:

    class PathLibNotAvailable(object):
        def __init__(self, *_, **__):
            raise RuntimeError("Please upgrade to python>=3.6 to use path-like objs")

    PathLike = PathLibNotAvailable
    fspath = PathLibNotAvailable


try:
    from PIL.Image import Image
except ImportError:

    class Image(object):
        def __init__(self, *_, **__):
            raise RuntimeError("pillow package is required for this functionality")

        save = __init__


try:
    from appium.webdriver import WebElement as AppiumWebElement
except (ImportError, TypeError):
    # TypeError is related to Appium-Python-Client above 5+ where they use tuple[...] syntax
    class AppiumWebElement(object):
        def __init__(self, *_, **__):
            raise RuntimeError(
                "appium-python-client package is required for this functionality"
            )


try:
    from selenium.common.exceptions import StaleElementReferenceException
    from selenium.webdriver.common.by import By
    from selenium.webdriver.remote.webdriver import WebDriver
    from selenium.webdriver.remote.webelement import WebElement
    from selenium.webdriver.support.event_firing_webdriver import EventFiringWebElement
except ImportError:

    class SeleniumNotInstalled(object):
        def __init__(self, *_, **__):
            raise RuntimeError("selenium package is required for this functionality")

    By = SeleniumNotInstalled
    EventFiringWebElement = SeleniumNotInstalled
    StaleElementReferenceException = SeleniumNotInstalled
    WebDriver = SeleniumNotInstalled
    WebElement = SeleniumNotInstalled

try:
    from playwright.sync_api import ElementHandle
    from playwright.sync_api import Locator as PlaywrightLocator
    from playwright.sync_api import Page
except ImportError:

    class PlaywrightNotInstalled(object):
        def __init__(self, *_, **__):
            raise RuntimeError("playwright package is required for this functionality")

    Page = PlaywrightNotInstalled
    PlaywrightLocator = PlaywrightNotInstalled
    ElementHandle = PlaywrightNotInstalled



# GENERATED FILE #
from enum import Enum

from applitools.common.utils.general_utils import DeprecatedEnumVariant

__all__ = ("DynamicTextType",)


class DynamicTextType(Enum):
    TextField = "TextField"
    Number = "Number"
    Date = "Date"
    Link = "Link"
    Email = "Email"
    Currency = "Currency"



from __future__ import absolute_import, division, print_function

import logging
import platform
from enum import Enum
from os import getcwd
from threading import Lock
from typing import TYPE_CHECKING, Any, List, Optional, Text

from .__version__ import __version__
from .connection import USDKConnection
from .errors import USDKFailure
from .schema import demarshal_error
from .target import ImageTarget
from .test_results import TestResults

if TYPE_CHECKING:
    from typing import Tuple, Type, Union

    from .batch_close import _EnabledBatchClose  # noqa
    from .command_context import CommandContext
    from .ec_client_settings import ECClientSettings
    from .extract_text import OCRRegion, TextRegionSettings
    from .fluent.web_check_settings import WebCheckSettings
    from .locators import VisualLocatorSettings
    from .object_registry import ObjectRegistry
    from .optional_deps import WebDriver
    from .protocol import USDKProtocol
    from .selenium import Configuration
    from .utils.custom_types import ViewPort

logger = logging.getLogger(__name__)

Failure = USDKFailure  # backward compatibility with eyes-selenium==5.0.0


class ManagerType(Enum):
    UFG = "ufg"
    CLASSIC = "classic"


class CommandExecutor(object):
    @classmethod
    def create(cls, protocol, mask_log):
        # type: (Type[USDKProtocol], Optional[bool]) -> CommandExecutor
        commands = cls(protocol, USDKConnection.create(mask_log=mask_log))
        name, version = protocol.SDK_INFO
        default_agent_id = "eyes.sdk.python/{}".format(__version__)
        environment = collect_environment()
        environment["sdk"] = {"lang": "python", "name": name, "currentVersion": version}
        commands.make_core(default_agent_id, getcwd(), environment)
        return commands

    @classmethod
    def get_instance(cls, protocol, mask_log=None):
        # type: (Type[USDKProtocol], Optional[bool]) -> CommandExecutor
        with _instances_lock:
            if protocol in _instances:
                return _instances[protocol]
            else:
                return _instances.setdefault(
                    protocol, cls.create(protocol, mask_log=mask_log)
                )

    def __init__(self, protocol, connection):
        # type: (Type[USDKProtocol], USDKConnection) -> None
        self._protocol = protocol
        self._connection = connection

    def make_core(self, agent_id, cwd, environment):
        # type: (Text, Text, dict) -> None
        self._connection.notification(
            "Core.makeCore",
            {
                "agentId": agent_id,
                "cwd": cwd,
                "environment": environment,
                "spec": self._protocol.commands_or_kind(),
            },
        )

    def core_make_ec_client(self, ec_client_settings):
        # type: (ECClientSettings) -> Text
        context = self._protocol.context(self._connection)
        m = context.marshaller
        settings = m.marshal_ec_client_settings(ec_client_settings)
        return self._checked_command(
            context, "Core.getECClient", {"settings": settings}
        )

    def core_make_manager(
        self, manager_type, concurrency=None, legacy_concurrency=None, agent_id=None
    ):
        # type: (ManagerType, Optional[int], Optional[int], Optional[Text]) -> dict
        context = self._protocol.context(self._connection)
        settings = {}
        if concurrency is not None:
            settings["concurrency"] = concurrency
        if legacy_concurrency is not None:
            settings["legacyConcurrency"] = legacy_concurrency
        if agent_id is not None:
            settings["agentId"] = agent_id
        payload = {"type": manager_type.value, "settings": settings}
        return self._checked_command(context, "Core.makeManager", payload)

    def core_get_viewport_size(self, driver):
        # type: (WebDriver) -> dict
        context = self._protocol.context(self._connection)
        m = context.marshaller
        target = m.marshal_webdriver_ref(driver)
        return self._checked_command(
            context, "Core.getViewportSize", {"target": target}
        )

    def core_set_viewport_size(self, driver, size):
        # type: (WebDriver, ViewPort) -> None
        context = self._protocol.context(self._connection)
        m = context.marshaller
        target = m.marshal_webdriver_ref(driver)
        self._checked_command(
            context,
            "Core.setViewportSize",
            {"target": target, "size": m.marshal_viewport_size(size)},
        )

    def core_close_batch(self, close_batch_settings):
        # type: (_EnabledBatchClose) -> None
        context = self._protocol.context(self._connection)
        m = context.marshaller
        settings = []
        for batch_id in close_batch_settings._ids:  # noqa
            close_batch_settings.batch_id = batch_id
            settings.append(m.marshal_enabled_batch_close(close_batch_settings))
        self._checked_command(context, "Core.closeBatch", {"settings": settings})

    def core_delete_test(self, test_results):
        # type: (TestResults) -> None
        context = self._protocol.context(self._connection)
        m = context.marshaller
        settings = m.marshal_delete_test_settings(test_results)
        self._checked_command(context, "Core.deleteTest", {"settings": settings})

    def manager_open_eyes(
        self,
        object_registry,  # type: ObjectRegistry
        manager,  # type: dict
        target=None,  # type: Optional[WebDriver]
        config=None,  # type: Optional[Configuration]
    ):
        # type: (...) -> dict
        context = self._protocol.context(self._connection, object_registry)
        m = context.marshaller
        payload = {"manager": manager}
        if target is not None:
            payload["target"] = m.marshal_webdriver_ref(target)
        if config is not None:
            cnf = m.marshal_configuration(config)
            payload["settings"] = cnf["open"]
            payload["config"] = cnf
        return self._checked_command(context, "EyesManager.openEyes", payload)

    def manager_get_results(self, manager, raise_ex, remove_duplicate_tests, timeout):
        # type: (dict, bool, Optional[bool], float) -> List[dict]
        context = self._protocol.context(self._connection)
        settings = {"throwErr": raise_ex}
        if remove_duplicate_tests is not None:
            settings["removeDuplicateTests"] = remove_duplicate_tests

        return self._checked_command(
            context,
            "EyesManager.getResults",
            {"manager": manager, "settings": settings},
            wait_timeout=timeout,
        )

    def eyes_check(
        self,
        object_registry,  # type: ObjectRegistry
        eyes,  # type: dict
        target,  # type: Optional[ImageTarget]
        settings,  # type: WebCheckSettings
        config,  # type: Configuration
    ):
        # type: (...) -> dict
        context = self._protocol.context(self._connection, object_registry)
        m = context.marshaller
        payload = {
            "eyes": eyes,
            "settings": m.marshal_check_settings(settings),
            "config": m.marshal_configuration(config),
        }
        if target:
            payload["target"] = m.marshal_image_target(target)
        return self._checked_command(context, "Eyes.check", payload)

    def core_locate(
        self,
        target,  # type: Union[WebDriver, ImageTarget]
        settings,  # type: VisualLocatorSettings
        config,  # type: Configuration
    ):
        # type: (...) -> dict
        context = self._protocol.context(self._connection)
        m = context.marshaller
        payload = {
            "settings": m.marshal_locate_settings(settings),
            "config": m.marshal_configuration(config),
        }
        if isinstance(target, ImageTarget):
            payload["target"] = m.marshal_image_target(target)
        else:
            payload["target"] = m.marshal_webdriver_ref(target)
        return self._checked_command(context, "Core.locate", payload)

    def core_extract_text(
        self,
        target,  # type: Union[WebDriver, ImageTarget]
        settings,  # type: Tuple[OCRRegion]
        config,  # type: Configuration
    ):
        # type: (...) -> List[Text]
        context = self._protocol.context(self._connection)
        m = context.marshaller
        payload = {
            "settings": m.marshal_ocr_extract_settings(settings),
            "config": m.marshal_configuration(config),
        }
        if isinstance(target, ImageTarget):
            payload["target"] = m.marshal_image_target(target)
        else:
            payload["target"] = m.marshal_webdriver_ref(target)
        return self._checked_command(context, "Core.extractText", payload)

    def core_locate_text(
        self,
        target,  # type: Union[WebDriver, ImageTarget]
        settings,  # type: TextRegionSettings
        config,  # type: Configuration
    ):
        # type: (...) -> dict
        context = self._protocol.context(self._connection)
        m = context.marshaller
        payload = {
            "settings": m.marshal_ocr_search_settings(settings),
            "config": m.marshal_configuration(config),
        }
        if isinstance(target, ImageTarget):
            payload["target"] = m.marshal_image_target(target)
        else:
            payload["target"] = m.marshal_webdriver_ref(target)
        return self._checked_command(context, "Core.locateText", payload)

    def eyes_close(self, object_registry, eyes, throw_err, config):
        # type: (ObjectRegistry, dict, bool, Configuration) -> List[dict]
        context = self._protocol.context(self._connection, object_registry)
        m = context.marshaller
        payload = {
            "eyes": eyes,
            "settings": {"throwErr": throw_err},
            "config": m.marshal_configuration(config),
        }
        return self._checked_command(context, "Eyes.close", payload)

    def eyes_abort(self, object_registry, eyes):
        # type: (ObjectRegistry, dict) -> List[dict]
        context = self._protocol.context(self._connection, object_registry)
        return self._checked_command(context, "Eyes.abort", {"eyes": eyes})

    def eyes_get_results(self, eyes, throw_err):
        context = self._protocol.context(self._connection)
        payload = {
            "eyes": eyes,
            "settings": {"throwErr": throw_err},
        }
        return self._checked_command(context, "Eyes.getResults", payload)

    def server_get_info(self):
        # type: () -> dict
        context = self._protocol.context(self._connection)
        return self._checked_command(context, "Server.getInfo", {})

    def _checked_command(self, command_context, name, payload, wait_timeout=9 * 60):
        # type: (CommandContext, Text, dict, float) -> Optional[Any]
        response = self._connection.command(
            command_context, name, payload, wait_timeout
        )
        response_payload = response["payload"]
        _check_error(response_payload)
        return response_payload.get("result")


def _check_error(payload):
    # type: (dict) -> None
    error = payload.get("error")
    if error:
        usdk_error = demarshal_error(error)
        logger.error("Re-raising an error received from SDK server: %r", usdk_error)
        raise usdk_error


def collect_versions_of_packages(*packages):
    try:  # assume python>=3.9
        from importlib.metadata import PackageNotFoundError
    except ImportError:  # older python
        from pkg_resources import DistributionNotFound as PackageNotFoundError

    from applitools.common.__version__ import _version

    versions = {}
    for package in packages:
        try:
            versions[package] = _version(package)
        except PackageNotFoundError:
            pass  # skip not installed packages
    return versions


def collect_environment():
    versions = collect_versions_of_packages(
        "appium-python-client",
        "core-universal",
        "eyes-common",
        "eyes-images",
        "eyes-playwright",
        "eyes-robotframework",
        "eyes-selenium",
        "robotframework",
        "robotframework-appiumlibrary",
        "robotframework-seleniumlibrary",
        "selenium",
    )
    versions["python"] = platform.python_version()
    return {"versions": versions}


_instances = {}
_instances_lock = Lock()



from __future__ import absolute_import, division, print_function

from datetime import timedelta, tzinfo

__all__ = ("UTC",)


class _UtcTz(tzinfo):
    """
    A UTC timezone class which is tzinfo compliant.
    """

    _ZERO = timedelta(0)

    def utcoffset(self, dt):
        return _UtcTz._ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return _UtcTz._ZERO


UTC = _UtcTz()



from __future__ import absolute_import, division, print_function, unicode_literals

from typing import TYPE_CHECKING

from .command_executor import CommandExecutor, ManagerType
from .config import DEFAULT_ALL_TEST_RESULTS_TIMEOUT
from .connection import TimeoutError
from .errors import DiffsFoundError, EyesError, NewTestError, TestFailedError
from .protocol import USDKProtocol
from .schema import demarshal_close_manager_results, demarshal_server_info
from .test_results import TestResultsSummary

if TYPE_CHECKING:
    from typing import Optional, Union

    from . import Configuration


class RunnerOptions(object):
    concurrency = 5
    _mask_log = False
    _remove_duplicate_tests = None

    def test_concurrency(self, value):
        # type: (int) -> RunnerOptions
        self.concurrency = value
        return self

    def remove_duplicate_tests(self, should_remove=True):
        # type: (bool) -> RunnerOptions
        self._remove_duplicate_tests = should_remove
        return self

    def mask_log(self, mask_log=True):
        self._mask_log = mask_log
        return self


class EyesRunner(object):
    AUTO_CLOSE_MODE_SYNC = True
    BASE_AGENT_ID = "eyes.sdk.python"
    Protocol = USDKProtocol

    def __init__(
        self,
        manager_type,  # type: ManagerType
        concurrency=None,  # type: Optional[int]
        is_legacy=None,  # type: Optional[bool]
        remove_duplicate_tests=None,  # type: Optional[bool]
        mask_log=None,  # type: Optional[bool]
    ):
        # type: (...) -> None
        self._connection_configuration = None
        self._remove_duplicate_tests = remove_duplicate_tests
        self._commands = CommandExecutor.get_instance(self.Protocol, mask_log)
        _, sdk_version = self.Protocol.SDK_INFO
        agent_id = "{}/{}".format(self.BASE_AGENT_ID, sdk_version)
        if is_legacy:
            self._ref = self._commands.core_make_manager(
                manager_type,
                legacy_concurrency=concurrency,
                agent_id=agent_id,
            )
        else:
            self._ref = self._commands.core_make_manager(
                manager_type,
                concurrency=concurrency,
                agent_id=agent_id,
            )

    def set_remove_duplicate_tests(self, should_remove):
        # type: (bool) -> None
        self._remove_duplicate_tests = should_remove

    @classmethod
    def get_server_info(cls):
        cmd = CommandExecutor.get_instance(cls.Protocol)
        result = cmd.server_get_info()
        return demarshal_server_info(result)

    def get_all_test_results(
        self, should_raise_exception=True, timeout=DEFAULT_ALL_TEST_RESULTS_TIMEOUT
    ):
        # type: (bool, Optional[int]) -> TestResultsSummary
        try:
            # Do not pass should_raise_exception because USDK raises untyped exceptions
            results = self._commands.manager_get_results(
                self._ref, should_raise_exception, self._remove_duplicate_tests, timeout
            )
        except TimeoutError:
            raise EyesError("Tests didn't finish in {} seconds".format(timeout))
        # We don't have server_url, api_key and proxy settings in runner
        # USDK should return them back as a part of TestResults
        structured_results = demarshal_close_manager_results(
            results, self._connection_configuration
        )
        for r in structured_results:
            if r.exception is not None:
                print("--- Test error. \n\tServer exception {}".format(r.exception))
            else:
                log_session_results_and_raise_exception(False, r.test_results)
        return structured_results

    def _set_connection_config(self, config):
        # type: (Configuration) -> None
        if self._connection_configuration is None:
            self._connection_configuration = config


class ClassicEyesRunner(EyesRunner):
    def __init__(self, options=RunnerOptions()):
        remove_duplicate_tests = options._remove_duplicate_tests  # noqa
        mask_log = options._mask_log  # noqa
        super(ClassicEyesRunner, self).__init__(
            ManagerType.CLASSIC,
            remove_duplicate_tests=remove_duplicate_tests,
            mask_log=mask_log,
        )


class VisualGridEyesRunner(EyesRunner):
    AUTO_CLOSE_MODE_SYNC = False

    def __init__(self, options_or_concurrency=RunnerOptions()):
        # type: (Union[RunnerOptions, int]) -> None
        if isinstance(options_or_concurrency, int):
            concurrency = options_or_concurrency * 5  # legacy factor
            remove_duplicate_tests = None
            is_legacy = True
            mask_log = None
        else:
            concurrency = options_or_concurrency.concurrency
            remove_duplicate_tests = (
                options_or_concurrency._remove_duplicate_tests  # noqa
            )
            mask_log = options_or_concurrency._mask_log  # noqa
            is_legacy = False
        super(VisualGridEyesRunner, self).__init__(
            ManagerType.UFG, concurrency, is_legacy, remove_duplicate_tests, mask_log
        )


def log_session_results_and_raise_exception(raise_ex, results):
    results_url = results.url
    scenario_id_or_name = results.name
    app_id_or_name = results.app_name
    if results.is_aborted:
        print("--- Test aborted.")
        if raise_ex:
            raise TestFailedError(results, scenario_id_or_name, app_id_or_name)
    elif results.is_unresolved:
        if results.is_new:
            print(
                "--- New test ended. \n\tPlease approve the new baseline at",
                results_url,
            )
            if raise_ex:
                raise NewTestError(results, scenario_id_or_name, app_id_or_name)
        else:
            print("--- Differences are found. \n\tSee details at", results_url)
            if raise_ex:
                raise DiffsFoundError(results, scenario_id_or_name, app_id_or_name)
    elif results.is_failed:
        print("--- Failed test ended. \n\tSee details at", results_url)
        if raise_ex:
            raise TestFailedError(results, scenario_id_or_name, app_id_or_name)
    else:
        print("--- Test passed. \n\tSee details at", results_url)



from __future__ import absolute_import, division, print_function

from typing import TYPE_CHECKING

from .marshaller import Marshaller

if TYPE_CHECKING:
    from typing import List, Union

    from .connection import USDKConnection
    from .object_registry import ObjectRegistry


class CommandContext(object):
    def __init__(self, connection, object_registry):
        # type: (USDKConnection, ObjectRegistry) -> None
        self._connection = connection
        self.object_registry = object_registry
        self.key = self.object_registry.next_command_key()
        self.marshaller = Marshaller(self.object_registry)

    @staticmethod
    def commands_or_kind():
        # type: () -> Union[str, List[str]]
        return "webdriver"

    def execute_callback(self, command):
        # type: (dict) -> None
        raise NotImplementedError



from __future__ import absolute_import, division, print_function

from typing import TYPE_CHECKING

import attr

if TYPE_CHECKING:
    from typing import Optional, Text

    from .config import ProxySettings


@attr.s
class ECClientCapabilitiesOptions(object):
    api_key = attr.ib(default=None)  # type: Optional[Text]
    server_url = attr.ib(default=None)  # type: Optional[Text]


@attr.s
class ECClientSettings(object):
    options = attr.ib(
        type=ECClientCapabilitiesOptions
    )  # type: ECClientCapabilitiesOptions
    proxy = attr.ib(default=None)  # type: Optional[ProxySettings]



# GENERATED FILE #
from enum import Enum

from applitools.common.utils.general_utils import DeprecatedEnumVariant

__all__ = ("IosMultiDeviceTarget",)


class IosMultiDeviceTarget(Enum):
    def __call__(self):
        from ._device_target import IosDeviceTarget

        return IosDeviceTarget(self.value)

    iPhone_SE_1 = "iPhone SE (1st generation)"
    iPhone_SE_2 = "iPhone SE (2nd generation)"
    iPhone_SE_3 = "iPhone SE (3rd generation)"
    iPhone_6 = "iPhone 6"
    iPhone_6_Plus = "iPhone 6 Plus"
    iPhone_7 = "iPhone 7"
    iPhone_7_Plus = "iPhone 7 Plus"
    iPhone_8 = "iPhone 8"
    iPhone_8_Plus = "iPhone 8 Plus"
    iPhone_X = "iPhone X"
    iPhone_XR = "iPhone XR"
    iPhone_Xs = "iPhone Xs"
    iPhone_Xs_Max = "iPhone Xs Max"
    iPhone_11 = "iPhone 11"
    iPhone_11_Pro = "iPhone 11 Pro"
    iPhone_11_Pro_Max = "iPhone 11 Pro Max"
    iPhone_12 = "iPhone 12"
    iPhone_12_mini = "iPhone 12 mini"
    iPhone_12_Pro = "iPhone 12 Pro"
    iPhone_12_Pro_Max = "iPhone 12 Pro Max"
    iPhone_13 = "iPhone 13"
    iPhone_13_mini = "iPhone 13 mini"
    iPhone_13_Pro = "iPhone 13 Pro"
    iPhone_13_Pro_Max = "iPhone 13 Pro Max"
    iPhone_14 = "iPhone 14"
    iPhone_14_Pro = "iPhone 14 Pro"
    iPhone_14_Plus = "iPhone 14 Plus"
    iPhone_14_Pro_Max = "iPhone 14 Pro Max"
    iPhone_15 = "iPhone 15"
    iPhone_15_Pro = "iPhone 15 Pro"
    iPhone_15_Plus = "iPhone 15 Plus"
    iPhone_15_Pro_Max = "iPhone 15 Pro Max"
    iPhone_16 = "iPhone 16"
    iPhone_16e = "iPhone 16e"
    iPhone_16_Pro = "iPhone 16 Pro"
    iPhone_16_Plus = "iPhone 16 Plus"
    iPhone_16_Pro_Max = "iPhone 16 Pro Max"
    iPhone_17 = "iPhone 17"
    iPhone_17_Pro = "iPhone 17 Pro"
    iPhone_17_Pro_Max = "iPhone 17 Pro Max"
    iPhone_Air = "iPhone Air"



from __future__ import absolute_import, division, print_function

from .optional_deps import (
    AppiumWebElement,
    ElementHandle,
    EventFiringWebElement,
    PlaywrightLocator,
    WebElement,
)


def is_list_or_tuple(elm):
    return isinstance(elm, (list, tuple))


def is_webelement(elm):
    return isinstance(
        elm,
        (
            WebElement,
            AppiumWebElement,
            EventFiringWebElement,
            PlaywrightLocator,
            ElementHandle,
        ),
    )



# GENERATED FILE #
from enum import Enum

from applitools.common.utils.general_utils import DeprecatedEnumVariant

__all__ = ("DeviceName",)


class DeviceName(Enum):
    iPhone_4 = "iPhone 4"
    iPhone_SE_3rd_Gen = "iPhone SE 3rd Gen"
    iPhone_6_7_8 = "iPhone 6/7/8"
    iPhone_6_7_8_Plus = "iPhone 6/7/8 Plus"
    iPhone_X = "iPhone X"
    iPhone_13_Mini = "iPhone 13 Mini"
    iPhone_14_Plus = "iPhone 14 Plus"
    iPhone_14_Pro = "iPhone 14 Pro"
    iPhone_15_Pro = "iPhone 15 Pro"
    BlackBerry_Z30 = "BlackBerry Z30"
    Nexus_4 = "Nexus 4"
    Nexus_5 = "Nexus 5"
    Nexus_5X = "Nexus 5X"
    Nexus_6 = "Nexus 6"
    Nexus_6P = "Nexus 6P"
    Pixel_2 = "Pixel 2"
    Pixel_2_XL = "Pixel 2 XL"
    LG_Optimus_L70 = "LG Optimus L70"
    Nokia_N9 = "Nokia N9"
    Nokia_Lumia_520 = "Nokia Lumia 520"
    Microsoft_Lumia_550 = "Microsoft Lumia 550"
    Microsoft_Lumia_950 = "Microsoft Lumia 950"
    Galaxy_S3 = "Galaxy S3"
    Galaxy_S_III = "Galaxy S III"
    Galaxy_S5 = "Galaxy S5"
    Kindle_Fire_HDX = "Kindle Fire HDX"
    iPad_Mini = "iPad Mini"
    iPad = "iPad"
    iPad_Pro = "iPad Pro"
    iPad_8th_Gen = "iPad 8th Gen"
    iPad_10th_Gen = "iPad 10th Gen"
    iPad_Mini_4th_Gen = "iPad Mini 4th Gen"
    iPad_Mini_6th_Gen = "iPad Mini 6th Gen"
    Blackberry_PlayBook = "Blackberry PlayBook"
    Nexus_10 = "Nexus 10"
    Nexus_7 = "Nexus 7"
    Galaxy_Note_3 = "Galaxy Note 3"
    Galaxy_Note_II = "Galaxy Note II"
    Galaxy_Note_2 = "Galaxy Note 2"
    Galaxy_S20 = "Galaxy S20"
    Galaxy_S22 = "Galaxy S22"
    Galaxy_S21 = "Galaxy S21"
    Galaxy_S21_Ultra = "Galaxy S21 Ultra"
    Galaxy_S22_Ultra = "Galaxy S22 Ultra"
    Laptop_with_touch = "Laptop with touch"
    Laptop_with_HiDPI_screen = "Laptop with HiDPI screen"
    Laptop_with_MDPI_screen = "Laptop with MDPI screen"
    iPhone_XR = "iPhone XR"
    iPhone_XS_Max = "iPhone XS Max"
    iPhone_XS = "iPhone XS"
    Samsung_Galaxy_A5 = "Samsung Galaxy A5"
    Galaxy_A5 = "Galaxy A5"
    Samsung_Galaxy_S8 = "Samsung Galaxy S8"
    Galaxy_S8 = "Galaxy S8"
    LG_G6 = "LG G6"
    iPad_Air_2 = "iPad Air 2"
    iPad_6th_Gen = "iPad 6th Gen"
    iPhone_11 = "iPhone 11"
    iPhone_11_Pro_Max = "iPhone 11 Pro Max"
    iPhone_11_Pro = "iPhone 11 Pro"
    Galaxy_S10 = "Galaxy S10"
    Galaxy_S9_Plus = "Galaxy S9 Plus"
    Galaxy_S9 = "Galaxy S9"
    Galaxy_S10_Plus = "Galaxy S10 Plus"
    Galaxy_S8_Plus = "Galaxy S8 Plus"
    Galaxy_Note_10 = "Galaxy Note 10"
    Galaxy_Note_10_Plus = "Galaxy Note 10 Plus"
    Galaxy_Note_9 = "Galaxy Note 9"
    Galaxy_Note_8 = "Galaxy Note 8"
    Galaxy_Note_4 = "Galaxy Note 4"
    Galaxy_Tab_A_SM_T510 = "Galaxy Tab A SM-T510"
    Galaxy_Tab_A7_SM_T500 = "Galaxy Tab A7 SM-T500"
    Galaxy_Tab_S8 = "Galaxy Tab S8"
    Galaxy_Tab_A8 = "Galaxy Tab A8"
    Galaxy_S23 = "Galaxy S23"
    Galaxy_S23_Ultra = "Galaxy S23 Ultra"
    Galaxy_S25 = "Galaxy S25"
    Galaxy_S25_Ultra = "Galaxy S25 Ultra"
    Galaxy_A52s = "Galaxy A52s"
    Pixel_3 = "Pixel 3"
    Pixel_3_XL = "Pixel 3 XL"
    Pixel_4 = "Pixel 4"
    Pixel_4_XL = "Pixel 4 XL"
    Pixel_5 = "Pixel 5"
    Pixel_9 = "Pixel 9"
    iPad_7th_Gen = "iPad 7th Gen"
    OnePlus_7T = "OnePlus 7T"
    OnePlus_7T_Pro = "OnePlus 7T Pro"
    Galaxy_Tab_S7 = "Galaxy Tab S7"
    Sony_Xperia_10_II = "Sony Xperia 10 II"
    Huawei_Mate_50_Pro = "Huawei Mate 50 Pro"
    Huawei_Matepad_11 = "Huawei Matepad 11"
    Huawei_P30 = "Huawei P30"
    Xiaomi_Mi_A3 = "Xiaomi Mi A3"
    Huawei_P40 = "Huawei P40"
    Xiaomi_Redmi_Note_8 = "Xiaomi Redmi Note 8"
    Xiaomi_Mi_Poco_X3_Pro = "Xiaomi Mi Poco X3 Pro"
    Xiaomi_Poco_X3 = "Xiaomi Poco X3"
    iPhone_5_SE = "iPhone 5/SE"
    Pixel_7 = "Pixel 7"
    iPhone_15 = "iPhone 15"

    @DeprecatedEnumVariant
    def iPhone_5SE(self):
        """`iPhone_5SE` is deprecated. Use `iPhone_5_SE` instead"""
        return self.iPhone_5_SE

    @DeprecatedEnumVariant
    def iPhone_5S_E(self):
        """`iPhone_5S_E` is deprecated. Use `iPhone_5SE` instead"""
        return self.iPhone_5SE



from __future__ import absolute_import, division, print_function

from enum import Enum
from typing import TYPE_CHECKING

import attr

from .accessibility import AccessibilitySettings
from .geometry import AccessibilityRegion, DynamicRegion, Rectangle, Region
from .utils.general_utils import DeprecatedEnumVariant

if TYPE_CHECKING:
    from typing import List, Optional, Text, Union

    from . import Point

__all__ = (
    "MatchLevel",
    "MatchResult",
    "FloatingMatchSettings",
    "ExactMatchSettings",
    "ImageMatchSettings",
    "FloatingBounds",
)


class MatchLevel(Enum):
    """
    The extent in which two images match (or are expected to match).
    """

    # Images do not necessarily match.
    NONE = "None"
    # Images have the same layout.
    LAYOUT = "Layout"
    # Images have the same layout.
    LAYOUT1 = "Layout1"
    # Images have the same layout.
    LAYOUT2 = "Layout2"
    # Images have the same content.
    IGNORE_COLORS = "IgnoreColors"
    # Images are nearly identical.
    STRICT = "Strict"
    # Images are identical.
    EXACT = "Exact"
    # Images are dynamic
    DYNAMIC = "Dynamic"

    @DeprecatedEnumVariant
    def CONTENT(self):
        # type: () -> MatchLevel
        """The `CONTENT` option that is being used will soon be deprecated.
        Please change it to `IGNORE_COLORS`"""
        return self.IGNORE_COLORS


# Obsolete type, Eyes.check doesn't return it anymore, kept for compatibility
@attr.s
class MatchResult(object):
    as_expected = attr.ib(
        default=None,
    )  # type: Optional[bool]
    window_id = attr.ib(
        default=None,
    )  # type: Optional[Text]
    screenshot = attr.ib(
        default=None,
    )


@attr.s(eq=False)
class FloatingMatchSettings(Rectangle):
    _region = attr.ib()  # type: Region
    _bounds = attr.ib()  # type: FloatingBounds
    left = attr.ib(init=False)  # type: int
    top = attr.ib(init=False)  # type: int
    width = attr.ib(init=False)  # type: int
    height = attr.ib(init=False)  # type: int
    max_up_offset = attr.ib(init=False)  # type: int
    max_down_offset = attr.ib(init=False)  # type: int
    max_left_offset = attr.ib(init=False)  # type: int
    max_right_offset = attr.ib(init=False)  # type: int

    def __str__(self):
        return (
            "FloatingMatchSettings({left}, {top}, {width} x {height} "
            "{max_up_offset}, {max_down_offset}, {max_left_offset}, "
            "{max_right_offset}".format(
                left=self.left,
                top=self.top,
                width=self.width,
                height=self.height,
                max_up_offset=self.max_up_offset,
                max_down_offset=self.max_down_offset,
                max_left_offset=self.max_left_offset,
                max_right_offset=self.max_right_offset,
            )
        )

    def __eq__(self, other):
        # type: (FloatingMatchSettings) -> bool
        return (
            self.left == other.left
            and self.top == other.top
            and self.width == other.width
            and self.height == other.height
            and self.max_up_offset == self.max_up_offset
            and self.max_down_offset == self.max_down_offset
            and self.max_left_offset == self.max_left_offset
            and self.max_right_offset == self.max_right_offset
        )

    def __attrs_post_init__(self):
        self.left = self._region.left
        self.top = self._region.top
        self.width = self._region.width
        self.height = self._region.height
        self.max_up_offset = self._bounds.max_up_offset
        self.max_down_offset = self._bounds.max_down_offset
        self.max_left_offset = self._bounds.max_left_offset
        self.max_right_offset = self._bounds.max_right_offset

    def offset(self, location_or_dx, dy=None):  # noqa
        # type: (Union[Point, int], Optional[int]) -> FloatingMatchSettings
        r = super(FloatingMatchSettings, self).offset(location_or_dx, dy)
        return FloatingMatchSettings(r, self._bounds)


@attr.s
class ExactMatchSettings(object):
    """
    Encapsulates settings for the :py:class:`MatchLevel.EXACT`.
    """

    min_diff_intensity = attr.ib(default=0)  # type: int
    min_diff_width = attr.ib(default=0)  # type: int
    min_diff_height = attr.ib(default=0)  # type: int
    match_threshold = attr.ib(default=0)  # type: float

    @classmethod
    def create_from(cls, other):
        # type: (ExactMatchSettings) -> ExactMatchSettings
        return ExactMatchSettings(
            min_diff_intensity=other.min_diff_intensity,
            min_diff_width=other.min_diff_width,
            min_diff_height=other.min_diff_height,
            match_threshold=other.match_threshold,
        )


@attr.s
class ImageMatchSettings(object):
    """
    Encapsulates match settings for the a session.
    """

    match_level = attr.ib(default=None)  # type: Optional[MatchLevel]
    exact = attr.ib(
        default=None, type=ExactMatchSettings
    )  # type: Optional[ExactMatchSettings]
    ignore_caret = attr.ib(default=None)  # type:bool
    use_dom = attr.ib(default=None)  # type: Optional[bool]
    enable_patterns = attr.ib(default=None)  # type: Optional[bool]
    ignore_displacements = attr.ib(default=None)  # type:Optional[bool]
    ignore_regions = attr.ib(type=Region, factory=list)  # type: List[Region]
    layout_regions = attr.ib(type=Region, factory=list)  # type: List[Region]
    strict_regions = attr.ib(type=Region, factory=list)  # type: List[Region]
    content_regions = attr.ib(type=Region, factory=list)  # type: List[Region]
    floating_match_settings = attr.ib(
        type=Region, factory=list
    )  # type: List[FloatingMatchSettings]
    accessibility = attr.ib(
        type=AccessibilityRegion, factory=list
    )  # type: List[AccessibilityRegion]
    accessibility_settings = attr.ib(
        default=None, type=AccessibilitySettings
    )  # type: Optional[AccessibilitySettings]
    dynamic_regions = attr.ib(
        type=DynamicRegion, factory=list
    )  # type: List[DynamicRegion]

    @classmethod
    def create_from(cls, other):
        # type: (ImageMatchSettings) -> ImageMatchSettings
        img = ImageMatchSettings(
            match_level=other.match_level,
            exact=ExactMatchSettings.create_from(other.exact) if other.exact else None,
            use_dom=other.use_dom,
        )
        img.ignore_caret = other.ignore_caret
        img.ignore_regions = other.ignore_regions
        img.layout_regions = other.layout_regions
        img.strict_regions = other.strict_regions
        img.content_regions = other.content_regions
        img.floating_match_settings = other.floating_match_settings
        img.enable_patterns = other.enable_patterns
        img.ignore_displacements = other.ignore_displacements
        img.accessibility = other.accessibility
        img.accessibility_settings = other.accessibility_settings
        img.dynamic_regions = other.dynamic_regions

        return img


@attr.s(slots=True)
class FloatingBounds(object):
    """A floating bounds defined by max_left_offset, max_up_offset, max_right_offset
    and max_down_offset"""

    max_left_offset = attr.ib(default=0)  # type: int
    max_up_offset = attr.ib(default=0)  # type: int
    max_right_offset = attr.ib(default=0)  # type: int
    max_down_offset = attr.ib(default=0)  # type: int



from __future__ import absolute_import, division, print_function

from typing import TYPE_CHECKING

from .. import common
from . import (
    DeviceName,
    MatchLevel,
    ScreenOrientation,
    SessionType,
    StitchMode,
    extract_text,
)
from .accessibility import (
    AccessibilityGuidelinesVersion,
    AccessibilityLevel,
    AccessibilityRegionType,
    AccessibilityStatus,
)
from .mmallow import (
    EXCLUDE,
    BaseSchema,
    Boolean,
    DateTime,
    Dict,
    Float,
    Function,
    Integer,
    List,
    Nested,
    Raw,
    Schema,
    SchemaMeta,
    String,
    post_dump,
    post_load,
    pre_load,
)
from .schema_fields import (
    BrowserInfo,
    DebugScreenshots,
    ElementReference,
    Enum,
    EnvironmentRaw,
    Error,
    FrameReference,
    LayoutBreakpoints,
    MobileOptions,
    NormalizationRaw,
    RegionReference,
    StitchOverlap,
    TargetReference,
    VisualGridOptions,
    demarshal_error,
)
from .selenium.misc import BrowserType
from .test_results import TestResultsStatus

if TYPE_CHECKING:
    import typing as t

    from . import config


# Default marshmallow.Schema has no option to skip attributes with None value
# or empty lists / dicts. Because it uses metaclass, it should be re-defined
# instead of simple subclassing
class USDKSchema(BaseSchema, metaclass=SchemaMeta):
    # Mimic marshmallow 2 behaviour
    class Meta:
        unknown = EXCLUDE

    __doc__ = BaseSchema.__doc__
    _always_skip_values = (None, [])
    _keep_empty_objects = ("lazyLoad",)  # fields that are allowed to have {} value

    @classmethod
    def should_keep(cls, key, value):
        # type: (t.Text, t.Any) -> bool
        if value in cls._always_skip_values:
            return False
        if value == {} and key not in cls._keep_empty_objects:
            return False
        return True

    @pre_load
    def _handle_aliases(self, data, **kwargs):
        """
        For each field that uses a data_key (alias), if the alias key is missing in the input
        but the original field name is present, copy the value over. This way the schema
        accepts both the alias key and the field name on load.
        """
        for field_name, field in self.fields.items():
            alias = field.data_key
            if alias and alias not in data and field_name in data:
                data[alias] = data[field_name]
        return data

    @post_dump
    def remove_none_values_empty_lists(self, data, **_):
        # type: (dict, **t.Any) -> dict
        return {k: v for k, v in data.items() if self.should_keep(k, v)}


class Size(USDKSchema):
    width = Integer()
    height = Integer()


class DebugScreenshotHandler(USDKSchema):
    debug_screenshots_path = String(data_key="path")
    debug_screenshots_prefix = String(data_key="prefix")


class Environment(USDKSchema):
    host_os = String(data_key="os")
    host_app = String(data_key="hostingApp")
    viewport_size = Nested(Size, data_key="viewportSize")


class EnvironmentRenderer(USDKSchema):
    render_environment_id = String(data_key="renderEnvironmentId")
    ec_session_id = String(data_key="ecSessionId")
    os = String()
    os_info = String(data_key="osInfo")
    hosting_app = String(data_key="hostingApp")
    hosting_app_info = String(data_key="hostingAppInfo")
    device_name = String(data_key="deviceName")
    viewport_size = Nested(Size, data_key="viewportSize")
    user_agent = String(data_key="userAgent")
    renderer = Dict()
    raw_environment = Dict(data_key="rawEnvironment")
    properties = List(Dict())

    @post_load
    def to_python(self, data, **_):
        return common.EnvironmentInfo(**data)


class DesktopBrowserRenderer(USDKSchema):
    width = Float()
    height = Float()
    browser_type = Enum(BrowserType, data_key="name")

    @post_load
    def to_python(self, data, **_):
        return common.DesktopBrowserInfo(**data)


class ChromeEmulationRenderer(USDKSchema):
    device_name = Enum(DeviceName, data_key="deviceName")
    screen_orientation = Enum(ScreenOrientation, data_key="screenOrientation")

    @post_load
    def to_python(self, data, **_):
        return common.ChromeEmulationInfo(**data)


class IosDeviceRenderer(USDKSchema):
    device_name = Function(
        data_key="deviceName",
        serialize=lambda obj, *args, **kwargs: getattr(
            obj.device_name, "value", obj.device_name
        ),
        deserialize=lambda value, *args, **kwargs: value,
    )
    screen_orientation = Enum(ScreenOrientation, data_key="screenOrientation")
    ios_version = Function(
        data_key="version",
        serialize=lambda obj, *args, **kwargs: obj.ios_version.value
        if obj.ios_version
        else None,
        deserialize=lambda value, *args, **kwargs: value,
    )

    @post_load
    def to_python(self, data, **_):
        return common.IosDeviceInfo(**data)


class AndroidDeviceRenderer(USDKSchema):
    device_name = Function(
        data_key="deviceName",
        serialize=lambda obj, *args, **kwargs: getattr(
            obj.device_name, "value", obj.device_name
        ),
        deserialize=lambda value, *args, **kwargs: value,
    )
    screen_orientation = Enum(ScreenOrientation, data_key="screenOrientation")
    android_version = Function(
        data_key="version",
        serialize=lambda obj, *args, **kwargs: obj.android_version.value
        if obj.android_version
        else None,
        deserialize=lambda value, *args, **kwargs: value,
    )

    @post_load
    def to_python(self, data, **_):
        return common.AndroidDeviceInfo(**data)


class Region(USDKSchema):
    left = Float(data_key="x")
    top = Float(data_key="y")
    width = Float()
    height = Float()

    @post_load
    def to_python(self, data, **_):
        return common.geometry.Region.from_(data)


class Offset(USDKSchema):
    max_left_offset = Float(data_key="left")
    max_up_offset = Float(data_key="top")
    max_right_offset = Float(data_key="right")
    max_down_offset = Float(data_key="bottom")


class ContextReference(USDKSchema):
    frame = FrameReference()
    scroll_root_locator = ElementReference(data_key="scrollRootElement")


class ImageCropRect(USDKSchema):
    header = Float(data_key="top")
    right = Float()
    footer = Float(data_key="bottom")
    left = Float()


class Normalization(USDKSchema):
    cut_provider = Nested(ImageCropRect, data_key="cut")
    rotation = Integer()
    scale_ratio = Float(data_key="scaleRatio")


class Batch(USDKSchema):
    id = String()
    name = String()
    sequence_name = String(data_key="sequenceName")
    started_at = DateTime("%Y-%m-%dT%H:%M:%SZ", data_key="startedAt")
    notify_on_completion = Boolean(data_key="notifyOnCompletion")
    properties = List(Dict())


class LayoutBreakpointsOptions(USDKSchema):
    breakpoints = Raw()
    reload = Boolean()
    height_breakpoints = Boolean(data_key="heightBreakpoints")


class Proxy(USDKSchema):
    url = String(required=True)
    username = String()
    password = String()


class ImageTarget(USDKSchema):
    image = String()
    dom = String()


class AccessibilitySettings(USDKSchema):
    level = Enum(AccessibilityLevel)
    guidelines_version = Enum(AccessibilityGuidelinesVersion, data_key="version")


class CodedRegionReference(USDKSchema):
    region = RegionReference()
    padding = Raw()
    region_id = String(data_key="regionId")


class FloatingRegionReference(USDKSchema):
    _target_path = RegionReference(data_key="region")
    _bounds = Nested(Offset(), data_key="offset")


class AccessibilityRegionReference(USDKSchema):
    _target_path = RegionReference(data_key="region")
    _type = Enum(AccessibilityRegionType, data_key="type")


class DynamicSettingsReference(Dict):
    ignore_patterns = List(String(), data_key="ignorePatterns")

    def _serialize(self, obj, *args, **kwargs):
        return None if obj is None else obj.to_dict()["ignore_patterns"]


class DynamicRegionReference(USDKSchema):
    _target_path = RegionReference(data_key="region")
    _dynamic_settings = DynamicSettingsReference(data_key="type")
    padding = Raw()


class LazyLoadOptions(USDKSchema):
    scroll_length = Integer(data_key="scrollLength")
    waiting_time = Integer(data_key="waitingTime")
    max_amount_to_scroll = Integer(data_key="maxAmountToScroll")


class EyesConfig(USDKSchema):
    # region
    # frames
    force_full_page_screenshot = Boolean(data_key="fully")
    # scrollRootElement
    stitch_mode = Enum(StitchMode, data_key="stitchMode")
    hide_scrollbars = Boolean(data_key="hideScrollbars")
    hide_caret = Boolean(data_key="hideCaret")
    stitch_overlap = StitchOverlap(data_key="overlap")
    wait_before_capture = Integer(data_key="waitBeforeCapture")
    # lazyLoad
    ignore_displacements = Boolean(
        attribute="default_match_settings.ignore_displacements",
        data_key="ignoreDisplacements",
    )
    # name
    # pageId
    ignore_regions = List(
        Nested(CodedRegionReference),
        attribute="default_match_settings.ignore_regions",
        data_key="ignoreRegions",
    )
    layout_regions = List(
        Nested(CodedRegionReference),
        attribute="default_match_settings.layout_regions",
        data_key="layoutRegions",
    )

    strict_regions = List(
        Nested(CodedRegionReference),
        attribute="default_match_settings.strict_regions",
        data_key="strictRegions",
    )
    content_regions = List(
        Nested(CodedRegionReference),
        attribute="default_match_settings.content_regions",
        data_key="contentRegions",
    )
    floating_match_settings = List(
        Nested(FloatingRegionReference),
        attribute="default_match_settings.floating_match_settings",
        data_key="floatingRegions",
    )
    accessibility = List(
        Nested(AccessibilityRegionReference),
        attribute="default_match_settings.accessibility",
        data_key="accessibilityRegions",
    )
    accessibility_settings = Nested(
        AccessibilitySettings,
        attribute="default_match_settings.accessibility_settings",
        data_key="accessibilitySettings",
    )
    dynamic_regions = List(
        Nested(DynamicRegionReference),
        attribute="default_match_settings.dynamic_regions",
        data_key="dynamicRegions",
    )

    match_level = Enum(
        MatchLevel,
        attribute="default_match_settings.match_level",
        data_key="matchLevel",
    )
    send_dom = Boolean(data_key="sendDom")
    use_dom = Boolean(attribute="default_match_settings.use_dom", data_key="useDom")
    enable_patterns = Boolean(
        attribute="default_match_settings.enable_patterns", data_key="enablePatterns"
    )
    ignore_caret = Boolean(
        attribute="default_match_settings.ignore_caret", data_key="ignoreCaret"
    )
    # enablePatterns
    # ignoreCaret
    visual_grid_options = VisualGridOptions(data_key="ufgOptions")
    layout_breakpoints = LayoutBreakpoints(data_key="layoutBreakpoints")
    disable_browser_fetching = Boolean(data_key="disableBrowserFetching")
    match_timeout = Float(data_key="matchTimeout")
    browsers_info = List(BrowserInfo(), data_key="environments")
    # autProxy
    normalization = NormalizationRaw()
    debug_images = DebugScreenshots(data_key="debugImages")
    mobile_options = MobileOptions(data_key="mobileOptions")
    # wait_before_screenshots = Float(data_key="waitBeforeScreenshots")


class EyesServerSettings(USDKSchema):
    server_url = String(data_key="eyesServerUrl")
    api_key = String(data_key="apiKey")
    agent_id = String(data_key="agentId")
    proxy = Nested(Proxy)
    # useDnsCache


class OpenSettings(EyesServerSettings, USDKSchema):
    app_name = String(data_key="appName")
    test_name = String(data_key="testName")
    # displayName
    user_test_id = String(data_key="userTestId")
    session_type = Enum(SessionType, data_key="sessionType")
    properties = List(Dict())
    batch = Nested(Batch)
    dont_close_batches = Boolean(data_key="keepBatchOpen")
    environment = EnvironmentRaw()
    environment_name = String(data_key="environmentName")
    baseline_env_name = String(data_key="baselineEnvName")
    branch_name = String(data_key="branchName")
    parent_branch_name = String(data_key="parentBranchName")
    baseline_branch_name = String(data_key="baselineBranchName")
    is_disabled = Boolean(data_key="isDisabled")
    # compareWithParentBranch
    # gitBranchingTimestamp
    # ignoreGitBranching
    # ignoreBaseline
    save_diffs = Boolean(data_key="saveDiffs")
    # abortIdleTestTimeout
    _timeout = Integer(data_key="connectionTimeout")
    # removeSession
    # isFunctionalTest
    # isComponentTest
    # ufgServerUrl
    # fallbackBaselineId
    # latestCommitInfo
    # processId
    # removeDuplicateTests

    # UFG related
    disable_nml_url_cache = Boolean(data_key="disableBrokerUrlCache")


class CheckSettings(USDKSchema):
    name = String()
    disable_browser_fetching = Boolean(data_key="disableBrowserFetching")
    layout_breakpoints = Nested(LayoutBreakpointsOptions, data_key="layoutBreakpoints")
    visual_grid_options = VisualGridOptions(data_key="ufgOptions")
    script_hooks = Dict(data_key="hooks")
    page_id = String(data_key="pageId")
    variation_group_id = String(data_key="userCommandId")
    timeout = Integer(data_key="matchTimeout")
    wait_before_capture = Integer(data_key="waitBeforeCapture")
    lazy_load = Nested(LazyLoadOptions, data_key="lazyLoad")
    # ScreenshotSettings
    region = TargetReference()
    frame_chain = List(Nested(ContextReference), data_key="frames")
    scroll_root_locator = ElementReference(data_key="scrollRootElement")
    stitch_content = Boolean(data_key="fully")
    # MatchSettings
    match_level = Enum(MatchLevel, data_key="matchLevel")
    send_dom = Boolean(data_key="sendDom")
    use_dom = Boolean(data_key="useDom")
    enable_patterns = Boolean(data_key="enablePatterns")
    ignore_caret = Boolean(data_key="ignoreCaret")
    ignore_displacements = Boolean(data_key="ignoreDisplacements")
    ignore_regions = List(Nested(CodedRegionReference), data_key="ignoreRegions")
    layout_regions = List(Nested(CodedRegionReference), data_key="layoutRegions")
    strict_regions = List(Nested(CodedRegionReference), data_key="strictRegions")
    content_regions = List(Nested(CodedRegionReference), data_key="contentRegions")
    floating_regions = List(Nested(FloatingRegionReference), data_key="floatingRegions")
    accessibility_regions = List(
        Nested(AccessibilityRegionReference), data_key="accessibilityRegions"
    )
    dynamic_regions = List(Nested(DynamicRegionReference), data_key="dynamicRegions")
    webview = Raw()
    screenshot_mode = String(data_key="screenshotMode")


class LocateSettings(USDKSchema):
    names = List(String(), data_key="locatorNames")
    first_only = Boolean(data_key="firstOnly")


class OCRSearchSettings(USDKSchema):
    _patterns = List(String(), data_key="patterns")
    _ignore_case = Boolean(data_key="ignoreCase")
    _first_only = Boolean(data_key="firstOnly")
    _language = String(data_key="language")


class ExtractTextSettings(USDKSchema):
    target = RegionReference(data_key="region")
    _hint = String(data_key="hint")
    _min_match = Float(data_key="minMatch")
    _language = String(data_key="language")


class CloseSettings(USDKSchema):
    # raise_ex = Boolean(data_key="throwErr")  # not present in config
    save_new_tests = Boolean(data_key="updateBaselineIfNew")
    save_failed_tests = Boolean(data_key="updateBaselineIfDifferent")


class CloseBatchSettings(USDKSchema):
    batch_id = String(data_key="batchId")
    server_url = String(data_key="eyesServerUrl")
    api_key = String(data_key="apiKey")
    proxy = Nested(Proxy)


class DeleteTestSettings(USDKSchema):
    id = String(data_key="testId")
    batch_id = String(data_key="batchId")
    secret_token = String(data_key="secretToken")
    server_url = String(
        attribute="_connection_config.server_url", data_key="eyesServerUrl"
    )
    api_key = String(attribute="_connection_config.api_key", data_key="apiKey")
    proxy = Nested(Proxy, attribute="_connection_config.proxy")


class ECClientCapabilitiesOptions(USDKSchema):
    server_url = String(data_key="ecServerUrl")
    api_key = String(data_key="apiKey")


class ECClientSettings(USDKSchema):
    options = Nested(ECClientCapabilitiesOptions)
    proxy = Nested(Proxy)


# De-marshaling schema
# Mimic marshmallow 2 behaviour
class DeMarshalingSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    @pre_load
    def _handle_aliases(self, data, **kwargs):
        """
        For each field that uses a data_key (alias), if the alias key is missing in the input
        but the original field name is present, copy the value over. This way the schema
        accepts both the alias key and the field name on load.
        """
        for field_name, field in self.fields.items():
            alias = field.data_key
            if alias and alias not in data and field_name in data:
                data[alias] = data[field_name]
        return data


class RectangleSize(DeMarshalingSchema):
    width = Integer()
    height = Integer()

    @post_load
    def to_python(self, data, **_):
        return common.RectangleSize(**data)


class ServerInfo(DeMarshalingSchema):
    logs_dir = String(data_key="logsDir")

    @post_load
    def to_python(self, data, **_):
        return common.ServerInfo(**data)


class SessionUrls(DeMarshalingSchema):
    batch = String()
    session = String()

    @post_load
    def to_python(self, data, **_):
        return common.test_results.SessionUrls(**data)


class ApiUrls(DeMarshalingSchema):
    baseline_image = String(data_key="baselineImage")
    current_image = String(data_key="currentImage")
    diff_image = String(data_key="diffImage")
    checkpoint_image = String(data_key="checkpointImage")
    checkpoint_image_thumbnail = String(data_key="checkpointImageThumbnail")
    side_by_side_image = String(data_key="sideBySideImage")

    @post_load
    def to_python(self, data, **_):
        return common.test_results.StepInfo.ApiUrls(**data)


class AppUrls(DeMarshalingSchema):
    step = String(data_key="step")
    step_editor = String(data_key="stepEditor")

    @post_load
    def to_python(self, data, **_):
        return common.test_results.StepInfo.AppUrls(**data)


class SessionAccessibilityStatus(DeMarshalingSchema):
    level = Enum(AccessibilityLevel)
    version = Enum(AccessibilityGuidelinesVersion)
    status = Enum(AccessibilityStatus)

    @post_load
    def to_python(self, data, **_):
        return common.accessibility.SessionAccessibilityStatus(**data)


class StepInfo(DeMarshalingSchema):
    name = String()
    is_different = Boolean(data_key="isDifferent")
    has_baseline_image = Boolean(data_key="hasBaselineImage")
    has_current_image = Boolean(data_key="hasCurrentImage")
    has_checkpoint_image = Boolean(data_key="hasCheckpointImage")
    api_urls = Nested(ApiUrls, data_key="apiUrls")
    app_urls = Nested(AppUrls, data_key="appUrls")

    @post_load
    def to_python(self, data, **_):
        return common.test_results.StepInfo(**data)


class LocateTextResponse(DeMarshalingSchema):
    left = Integer(data_key="x")
    top = Integer(data_key="y")
    width = Integer()
    height = Integer()
    text = String()

    @post_load
    def to_python(self, data, **_):
        return extract_text.TextRegion(**data)


class TestResults(DeMarshalingSchema):
    steps = Integer()
    matches = Integer()
    mismatches = Integer()
    missing = Integer()
    exact_matches = Integer(data_key="exactMatches")
    strict_matches = Integer(data_key="strictMatches")
    content_matches = Integer(data_key="contentMatches")
    layout_matches = Integer(data_key="layoutMatches")
    none_matches = Integer(data_key="noneMatches")
    is_new = Boolean(data_key="isNew")
    url = String()
    status = Enum(TestResultsStatus)
    name = String()
    secret_token = String(data_key="secretToken")
    id = String()
    app_name = String(data_key="appName")
    batch_name = String(data_key="batchName")
    batch_id = String(data_key="batchId")
    branch_name = String(data_key="branchName")
    host_os = String(data_key="hostOS")
    host_app = String(data_key="hostApp")
    host_display_size = Nested(RectangleSize, data_key="hostDisplaySize")
    started_at = String(data_key="startedAt")
    duration = Integer()
    is_different = Boolean(data_key="isDifferent")
    is_aborted = Boolean(data_key="isAborted")
    is_empty = Boolean(data_key="isEmpty")
    app_urls = Nested(SessionUrls, data_key="appUrls")
    api_urls = Nested(SessionUrls, data_key="apiUrls")
    steps_info = List(Nested(StepInfo), data_key="stepsInfo")
    baseline_id = String(data_key="baselineId")
    accessibility_status = Nested(
        SessionAccessibilityStatus, data_key="accessibilityStatus"
    )
    user_test_id = String(data_key="userTestId")

    @post_load
    def to_python(self, data, **_):
        return common.TestResults(**data)


class TestResultContainer(DeMarshalingSchema):
    test_results = Nested(TestResults, data_key="result")
    browser_info = BrowserInfo(data_key="environment")
    exception = Error(allow_none=True, data_key="error")
    user_test_id = String(data_key="userTestId")

    @post_load
    def to_python(self, data, **_):
        return common.TestResultContainer(**data)


class TestResultsSummary(DeMarshalingSchema):
    results = List(Nested(TestResultContainer))
    exceptions = Integer()
    passed = Integer()
    unresolved = Integer()
    failed = Integer()
    # these attributes get None value when Eyes.locate call fails
    mismatches = Integer(allow_none=True)
    missing = Integer(allow_none=True)
    matches = Integer(allow_none=True)

    @post_load
    def to_python(self, data, **_):
        return common.TestResultsSummary(**data)


def demarshal_locate_result(results):
    # type: (dict) -> t.Dict[t.Text, t.List[common.Region]]
    return {
        locator_id: [Region().load(r) for r in regions] if regions else []
        for locator_id, regions in results.items()
    }


def demarshal_locate_text_result(results):
    # type: (dict) -> t.Dict[t.Text, t.List[extract_text.TextRegion]]
    return {
        locator_id: ([LocateTextResponse().load(r) for r in regions] if regions else [])
        for locator_id, regions in results.items()
    }


def demarshal_test_results(results_list, conf):
    # type: (t.List[dict], config.Configuration) -> t.List[common.TestResults]
    # When locating visual locators, result might be None
    results = [TestResults().load(r) for r in results_list if r]
    for result in results:
        result.set_connection_config(conf.server_url, conf.api_key, conf.proxy)
    return results


def demarshal_close_manager_results(close_manager_result_dict, conf):
    # type: (dict, config.Configuration) -> common.TestResultsSummary
    results = TestResultsSummary().load(close_manager_result_dict)
    for container in results:
        if container.test_results:
            container.test_results.set_connection_config(
                conf.server_url, conf.api_key, conf.proxy
            )
    return results


def demarshal_server_info(info_dict):
    # type: (dict) -> common.ServerInfo
    return ServerInfo().load(info_dict)



from __future__ import absolute_import, division, print_function

from copy import copy
from typing import Dict, List, Text, Union

import attr

from .geometry import Region
from .utils import argument_guard
from .validators import is_list_or_tuple

__all__ = ("VisualLocator", "VisualLocatorSettings")

LOCATORS_TYPE = Dict[Text, List[Region]]  # typedef


def _clean_names(names):
    argument_guard.not_list_or_tuple(names)
    if len(names) == 1 and is_list_or_tuple(names[0]):
        names = names[0]
    argument_guard.are_(names, klass=str)
    return names


@attr.s
class VisualLocatorSettingsValues(object):
    names = attr.ib()  # type: List[Text]
    first_only = attr.ib(default=True)  # type: bool
    image = attr.ib(default=None)  # type: Text | None

    @property
    def is_first_only(self):
        # type: () -> bool
        return self.first_only


class VisualLocatorSettings(object):
    def __init__(self, *names):
        # type: (*Text) -> None
        self.values = VisualLocatorSettingsValues(names=list(names))

    def first(self):
        # type: () -> VisualLocatorSettings
        clone = self.clone()
        clone.values.first_only = True
        return clone

    def all(self):
        # type: () -> VisualLocatorSettings
        clone = self.clone()
        clone.values.first_only = False
        return clone

    def clone(self):
        # type: () -> VisualLocatorSettings
        return copy(self)

    def name(self, name):
        # type: (Text) -> VisualLocatorSettings
        argument_guard.is_a(name, str)
        cloned = self.clone()
        cloned.values.names.append(name)
        return cloned

    def names(self, *names):
        # type: (*Union[Text, List[Text]]) -> VisualLocatorSettings
        cleaned_names = _clean_names(names)
        cloned = self.clone()
        cloned.values.names = cleaned_names
        return cloned

    def image(self, image):
        # type: (Text) -> VisualLocatorSettings
        argument_guard.is_a(image, str)
        cloned = self.clone()
        cloned.values.image = image
        return cloned


class VisualLocator(object):
    @staticmethod
    def name(name):
        # type: (Text) -> VisualLocatorSettings
        argument_guard.is_a(name, str)
        return VisualLocatorSettings(name)

    @staticmethod
    def names(*names):
        # type: (*Union[Text, List[Text]]) -> VisualLocatorSettings
        return VisualLocatorSettings(*_clean_names(names))


@attr.s
class VisualLocatorsData(object):
    app_name = attr.ib()
    image_url = attr.ib()
    first_only = attr.ib()
    locator_names = attr.ib()



from __future__ import absolute_import, division, print_function

import math
from abc import ABCMeta, abstractmethod

import attr

__all__ = ("FixedCutProvider", "UnscaledFixedCutProvider", "NullCutProvider")


@attr.s
class CutProvider(object, metaclass=ABCMeta):
    header = attr.ib()  # type: int
    footer = attr.ib()  # type: int
    left = attr.ib()  # type: int
    right = attr.ib()  # type: int

    @abstractmethod
    def scale(self, scale_ratio):
        # type: (float) -> CutProvider
        """Get a scaled version of the cut provider"""


class FixedCutProvider(CutProvider):
    def scale(self, scale_ratio):
        # type: (float) -> FixedCutProvider
        cut_provider = FixedCutProvider(
            math.ceil(self.header * scale_ratio),
            math.ceil(self.footer * scale_ratio),
            math.ceil(self.left * scale_ratio),
            math.ceil(self.right * scale_ratio),
        )
        return cut_provider


class UnscaledFixedCutProvider(CutProvider):
    def scale(self, scale_ratio):
        # type: (float) -> UnscaledFixedCutProvider
        if isinstance(self, NullCutProvider):
            return self

        cut_provider = UnscaledFixedCutProvider(
            self.header, self.footer, self.left, self.right
        )
        return cut_provider


class NullCutProvider(UnscaledFixedCutProvider):
    def __init__(self):
        super(NullCutProvider, self).__init__(0, 0, 0, 0)



# GENERATED FILE #
from enum import Enum

from applitools.common.utils.general_utils import DeprecatedEnumVariant

__all__ = ("AndroidMultiDeviceTarget",)


class AndroidMultiDeviceTarget(Enum):
    def __call__(self):
        from ._device_target import AndroidDeviceTarget

        return AndroidDeviceTarget(self.value)

    Galaxy_S25 = "Galaxy S25"
    Galaxy_S25_Ultra = "Galaxy S25 Ultra"
    Pixel_9 = "Pixel 9"



from __future__ import absolute_import, division, print_function

from .__version__ import __version__  # noqa
from ._dynamic_regions import DynamicSettings, DynamicTextType
from .accessibility import (
    AccessibilityGuidelinesVersion,
    AccessibilityLevel,
    AccessibilityRegionType,
    AccessibilitySettings,
    AccessibilityStatus,
)
from .config import BatchInfo, Configuration, ProxySettings
from .errors import (
    DiffsFoundError,
    EyesError,
    NewTestError,
    OutOfBoundsError,
    TestFailedError,
    USDKFailure,
)
from .geometry import (
    AccessibilityRegion,
    CoordinatesType,
    Point,
    RectangleSize,
    Region,
    SubregionForStitching,
)
from .layout_breakpoints_options import LayoutBreakpointsOptions
from .logger import FileLogger, StdoutLogger
from .match import (
    ExactMatchSettings,
    FloatingBounds,
    FloatingMatchSettings,
    ImageMatchSettings,
    MatchLevel,
    MatchResult,
)
from .selenium.misc import MobileOptions, StitchMode
from .server import FailureReports, ServerInfo, SessionType
from .test_results import (
    TestResultContainer,
    TestResults,
    TestResultsStatus,
    TestResultsSummary,
)
from .ultrafastgrid.config import (
    DeviceName,
    IosDeviceName,
    IosVersion,
    ScreenOrientation,
    VisualGridOption,
)
from .ultrafastgrid.render_browser_info import (
    AndroidDeviceInfo,
    ChromeEmulationInfo,
    DesktopBrowserInfo,
    EnvironmentInfo,
    IosDeviceInfo,
)

__all__ = (
    "AccessibilityGuidelinesVersion",
    "AccessibilityLevel",
    "AccessibilityRegion",
    "AccessibilityRegionType",
    "AccessibilitySettings",
    "AccessibilityStatus",
    "ChromeEmulationInfo",
    "CoordinatesType",
    "DesktopBrowserInfo",
    "DeviceName",
    "DiffsFoundError",
    "DynamicTextType",
    "EnvironmentInfo",
    "ExactMatchSettings",
    "EyesError",
    "FailureReports",
    "FileLogger",
    "FloatingBounds",
    "FloatingMatchSettings",
    "ImageMatchSettings",
    "LayoutBreakpointsOptions",
    "MatchLevel",
    "MatchResult",
    "NewTestError",
    "OutOfBoundsError",
    "Point",
    "RectangleSize",
    "Region",
    "ScreenOrientation",
    "ServerInfo",
    "SessionType",
    "StdoutLogger",
    "StitchMode",
    "SubregionForStitching",
    "TestFailedError",
    "TestResultContainer",
    "TestResults",
    "TestResultsStatus",
    "TestResultsSummary",
    "USDKFailure",
    "logger",
    "MobileOptions",
    "IosDeviceInfo",
    "AndroidDeviceInfo",
)



from __future__ import absolute_import, division, print_function

from typing import TYPE_CHECKING

import attr

if TYPE_CHECKING:
    from typing import List, Optional, Union


@attr.s
class LayoutBreakpointsOptions(object):
    breakpoints = attr.ib()  # type: Union[bool, List[int]]
    reload = attr.ib(default=None)  # type: Optional[bool]
    height_breakpoints = attr.ib(default=None)  # type: Optional[bool]



from __future__ import absolute_import, division, print_function


class LooseVersion(object):
    """Minimal replacement for deprecated distutils.version.LooseVersion"""

    def __init__(self, version):
        # type: (str) -> None
        self.version = tuple(int(c) for c in version.split("."))



from __future__ import absolute_import, division, print_function

from typing import TYPE_CHECKING, overload

from .images_check_settings import ImagesCheckSettings, image_path_or_bytes

if TYPE_CHECKING:
    from typing import ByteString, Text, Union

    from ..geometry import Region
    from ..optional_deps import Image, PathLike


class Target(object):
    """
    Fluent interface factory class for image targets
    """

    @staticmethod  # noqa
    @overload
    def image(image):
        # type: (Image) -> ImagesCheckSettings
        pass

    @staticmethod  # noqa
    @overload
    def image(image):
        # type: (ByteString) -> ImagesCheckSettings
        pass

    @staticmethod  # noqa
    @overload
    def image(path):
        # type: (Union[Text, PathLike]) -> ImagesCheckSettings
        pass

    @staticmethod  # noqa
    def image(image_or_path):
        check_settings = ImagesCheckSettings()
        check_settings.values.image = image_path_or_bytes(image_or_path)
        return check_settings

    @staticmethod  # noqa
    @overload
    def region(image, rect):
        # type: (Image, Region) -> ImagesCheckSettings
        pass

    @staticmethod  # noqa
    @overload
    def region(image, rect):
        # type: (ByteString, Region) -> ImagesCheckSettings
        pass

    @staticmethod  # noqa
    @overload
    def region(path, rect):
        # type: (Union[Text, PathLike], Region) -> ImagesCheckSettings
        pass

    @staticmethod  # noqa
    def region(image_or_path, rect):
        check_settings = Target.image(image_or_path)
        check_settings.values.target_region = rect
        return check_settings



from __future__ import absolute_import, division, print_function, unicode_literals

from ..common import deprecated
from ..common.locators import VisualLocator, VisualLocatorSettings

deprecated.module(__name__)
__all__ = "VisualLocator", "VisualLocatorSettings"



from __future__ import absolute_import, division, print_function

from itertools import count
from typing import TYPE_CHECKING
from uuid import uuid1

if TYPE_CHECKING:
    from typing import Text


class RefId(object):
    def __init__(self, object_registry_id, obj_id):
        # type: (Text, Text) -> None
        self.object_registry_id = object_registry_id
        self.obj_id = obj_id

    def __str__(self):
        return "{}:{}".format(self.object_registry_id, self.obj_id)

    @classmethod
    def from_str(cls, val):
        # type: (Text) -> RefId
        command_key, obj_id = val.rsplit(":", 1)
        return cls(command_key, obj_id)


class ObjectRegistry(object):
    def __init__(self):
        # type: () -> None
        self.id = str(uuid1())
        self._command_key_gen = count(1)

    def next_command_key(self):
        # type: () -> RefId
        return RefId(self.id, str(next(self._command_key_gen)))

    def marshal_driver(self, driver):
        raise NotImplementedError

    def marshal_element(self, element):
        raise NotImplementedError
